package quimufu.colourful_portals;

import eu.midnightdust.lib.config.MidnightConfig;
import net.fabricmc.api.ModInitializer;
import net.fabricmc.fabric.api.event.lifecycle.v1.ServerLifecycleEvents;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.fabricmc.fabric.api.itemgroup.v1.FabricItemGroupEntries;
import net.fabricmc.fabric.api.itemgroup.v1.ItemGroupEvents;
import net.minecraft.block.AbstractBlock;
import net.minecraft.block.BlockState;
import net.minecraft.block.enums.NoteBlockInstrument;
import net.minecraft.entity.EntityType;
import net.minecraft.item.*;
import net.minecraft.registry.Registries;
import net.minecraft.registry.Registry;
import net.minecraft.registry.RegistryKey;
import net.minecraft.server.MinecraftServer;
import net.minecraft.sound.BlockSoundGroup;
import net.minecraft.util.Identifier;
import net.minecraft.util.Rarity;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.BlockView;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import quimufu.colourful_portals.config.ColourfulPortalConfig;
import quimufu.colourful_portals.portal.*;
import quimufu.colourful_portals.portal_fluid.PortalFluid;
import quimufu.colourful_portals.portal_fluid.PortalFluidBlock;
import quimufu.colourful_portals.portal_fluid.PortalFluidBucketItem;

import java.util.Comparator;
import java.util.HashSet;

import static quimufu.colourful_portals.Components.PORTAL_CANDIDATE_LIST;
import static quimufu.colourful_portals.Components.PORTAL_LIST;

public class ColourfulPortalsMod implements ModInitializer {

    public static final String MOD_ID = "colourful_portals";
    public static final Logger LOGGER = LoggerFactory.getLogger(MOD_ID);
    public static final HashSet<Identifier> PORTAL_BLOCKS = new HashSet<>();

    public static final PortalBlock PORTAL_BLOCK = new PortalBlock(AbstractBlock.Settings.create().instrument(NoteBlockInstrument.HAT).sounds(BlockSoundGroup.GLASS).strength(-1.0f, 3600000.8f).dropsNothing().nonOpaque().luminance((bs) -> 15).allowsSpawning(ColourfulPortalsMod::never).solidBlock(ColourfulPortalsMod::never).suffocates(ColourfulPortalsMod::never).blockVision(ColourfulPortalsMod::never).ticksRandomly());
    public static final BlockItem PORTAL_BLOCK_ITEM = new BlockItem(PORTAL_BLOCK, new Item.Settings().rarity(Rarity.EPIC));
    public static final Item BLOB_DARK = new Item(new Item.Settings());
    public static final Item BLOB_BRIGHT = new Item(new Item.Settings());
    public static final PortalFluid PORTAL_FLUID = new PortalFluid();
    public static final PortalFluidBlock PORTAL_FLUID_BLOCk = new PortalFluidBlock(PORTAL_FLUID, AbstractBlock.Settings.create().sounds(BlockSoundGroup.INTENTIONALLY_EMPTY).luminance((bs) -> 15).noCollision().strength(100.0f).dropsNothing());
    public static final BucketItem PORTAL_FLUID_BUCKET_ITEM = new PortalFluidBucketItem(PORTAL_FLUID, new Item.Settings().recipeRemainder(Items.BUCKET).maxCount(1).rarity(Rarity.RARE));
    public static PortalManager PORTAL_MANAGER;

    public static final RegistryKey<Registry<PrioritizedPortalLinkingSystemBuilder>> PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY_KEY = RegistryKey.ofRegistry(Identifier.of(MOD_ID, "portal_linking_system"));
    public static final Registry<PrioritizedPortalLinkingSystemBuilder> PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY = FabricRegistryBuilder.createSimple(PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY_KEY).buildAndRegister();

    private static boolean never(BlockState blockState, BlockView blockView, BlockPos blockPos, EntityType<?> entityType) {
        return false;
    }

    private static boolean never(BlockState blockState, BlockView blockView, BlockPos blockPos) {
        return false;
    }


    @Override
    public void onInitialize() {
        LOGGER.info("Colouring Portals...");
        MidnightConfig.init(MOD_ID, ColourfulPortalConfig.class);

        for (String id : ColourfulPortalConfig.getAllPortalBlocks()) {
            LOGGER.info(id);
            PORTAL_BLOCKS.add(Identifier.tryParse(id));
        }
        if (!ColourfulPortalConfig.disableImmersivePortals && hasImmPtl()) {
            Registry.register(PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY, ImmersivePortalsLinkingSystem.IMMERSIVE_PORTALS_LINKING_SYSTEM, new PrioritizedPortalLinkingSystemBuilder(ImmersivePortalsLinkingSystem::new, 100));
        }
        Registry.register(PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY, DefaultLinkingSystem.DEFAULT_LINKING_SYSTEM, new PrioritizedPortalLinkingSystemBuilder(DefaultLinkingSystem::new, 90));

        Identifier identifier = Identifier.of(MOD_ID, "portal_block");

        Registry.register(Registries.BLOCK, identifier, PORTAL_BLOCK);
        Registry.register(Registries.ITEM, identifier, PORTAL_BLOCK_ITEM);

        Registry.register(Registries.ITEM, Identifier.of(MOD_ID, "portal_fluid_bucket"), PORTAL_FLUID_BUCKET_ITEM);
        Registry.register(Registries.FLUID, Identifier.of(MOD_ID, "portal_fluid"), PORTAL_FLUID);
        Registry.register(Registries.BLOCK, Identifier.of(MOD_ID, "portal_fluid_block"), PORTAL_FLUID_BLOCk);

        Registry.register(Registries.ITEM, Identifier.of(MOD_ID, "colour_blob_bright"), BLOB_BRIGHT);
        Registry.register(Registries.ITEM, Identifier.of(MOD_ID, "colour_blob_dark"), BLOB_DARK);

        ItemGroupEvents.modifyEntriesEvent(ItemGroups.INGREDIENTS)
                .register(ColourfulPortalsMod::addToIngredients);

        ItemGroupEvents.modifyEntriesEvent(ItemGroups.TOOLS)
                .register(ColourfulPortalsMod::addTtTools);

        ServerLifecycleEvents.SERVER_STARTED.register(this::onServerStarted);
        LOGGER.info("Portals Colourful!");
    }

    private boolean hasImmPtl() {
        try {
            Class.forName("qouteall.imm_ptl.core.IPModMain", false, this.getClass().getClassLoader());
        } catch (ClassNotFoundException e) {
            return false;
        }
        return true;
    }

    private void onServerStarted(MinecraftServer minecraftServer) {

        PortalListComponent portalCandidateList = PORTAL_CANDIDATE_LIST.get(minecraftServer.getSaveProperties().getMainWorldProperties());
        PortalListComponent portalList = PORTAL_LIST.get(minecraftServer.getSaveProperties().getMainWorldProperties());

        PortalLinkingSystem currentLinkingSystem = getCurrentlyConfiguredLinkingSystem(minecraftServer);
        if (currentLinkingSystem.getLinkingSystemId() != portalList.lastPortalLinkingSystemId() || currentLinkingSystem.needsReInit()) {
            PrioritizedPortalLinkingSystemBuilder prevSystem;
            if ((prevSystem = PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY.get(portalList.lastPortalLinkingSystemId())) != null) {
                PortalLinkingSystem prevLS = prevSystem.portalLinkingSystemBuilder().build(minecraftServer);
                for (Identifier blockId : portalList.getBlockIds()) {
                    for (PortalRepresentation portal : portalList.getPortals(blockId)) {
                        prevLS.unLinkPortal(portal);
                    }
                }
            }
            for (Identifier blockId : portalList.getBlockIds()) {
                currentLinkingSystem.linkPortals(portalList.getPortals(blockId));
            }
        }
        PORTAL_MANAGER = new PortalManager(currentLinkingSystem, portalCandidateList, portalList);
        PORTAL_MANAGER.onLoad(minecraftServer);
    }

    private PortalLinkingSystem getCurrentlyConfiguredLinkingSystem(MinecraftServer minecraftServer) {
        return PORTAL_LINKING_SYSTEM_BUILDER_REGISTRY
                .stream()
                .min(Comparator.comparing(PrioritizedPortalLinkingSystemBuilder::priority))
                .map(PrioritizedPortalLinkingSystemBuilder::portalLinkingSystemBuilder)
                .orElse(DefaultLinkingSystem::new)
                .build(minecraftServer);
    }

    private static void addTtTools(FabricItemGroupEntries entries) {
        entries.add(PORTAL_FLUID_BUCKET_ITEM);
    }

    private static void addToIngredients(FabricItemGroupEntries content) {
        content.add(BLOB_BRIGHT);
        content.add(BLOB_DARK);
    }

}